#! /usr/bin/env python
# -*- coding: utf-8 -*
import collections
from pprint import pprint
import census_data_aggregator
from census_data_downloader.core import MOE_MAP
from census_data_downloader.core.tables import BaseTableConfig
from census_data_downloader.core.decorators import register


@register
class InternetDownloader(BaseTableConfig):
    YEAR_LIST = (2017,)
    PROCESSED_TABLE_NAME = "internet"
    UNIVERSE = "households"
    RAW_TABLE_NAME = 'B28002'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': "universe",
        '002': "internet_any_source",
        '003': "dialup_only",
        '004': "broadband_any_source",
        '005': "cellular_data",
        '006': "cellular_data_only",
        '007': "broadband_cable_fiber_or_dsl",
        '008': "broadband_only",
        '009': "satellite",
        '010': "satellite_only",
        '011': "other_only",
        '012': "internet_without_subscription",
        '013': "no_internet"
    })

    def process(self, df):
        # This field, which combines people with no internet and those only only receive via
        # a free program like municipal wifi together into a combined group.
        # The Census Bureau considers this to be the true number of households without Internet access.
        # Calculate moe for these rows so we can use in the median aggregator
        def calculate_moe(row):
            pprint(row)
            if row['internet_without_subscription_moe'] in list(MOE_MAP.values()):
                value = sum([row['internet_without_subscription'], row['no_internet']])
                moe = None
            value, moe = census_data_aggregator.approximate_sum(
                (row['internet_without_subscription'], row['internet_without_subscription_moe']),
                (row['no_internet'], row['no_internet_moe']),
            )
            row['total_no_internet_and_no_subscription'] = value
            row['total_no_internet_and_no_subscription_moe'] = moe
            return row

        df = df.apply(
            calculate_moe,
            axis=1
        )

        # Pass it back
        return df



#! /usr/bin/env python
# -*- coding: utf-8 -*
import collections
from census_data_downloader.core.tables import BaseTableConfig
from census_data_downloader.core.decorators import register


@register
class VoterDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "voters"
    """ citizen voting-age population by age"""
    UNIVERSE = "citizens 18 years and over"
    RAW_TABLE_NAME = 'B29001'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        "001": "total",
        "002": "total_18_29_yrs_old",
        "003": "total_30_44_yrs_old",
        "004": "total_45_64_yrs_old",
        "005": "total_65_older"
    })



#! /usr/bin/env python
# -*- coding: utf-8 -*
import collections
from census_data_downloader.core.tables import BaseTableConfig
from census_data_downloader.core.decorators import register


@register
class HousingValueDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "housingvalue"
    UNIVERSE = "owner occupied housing units"
    RAW_TABLE_NAME = "B25075"
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        "001": "total",
        "002": "0_10000",
        "003": "10000_14999",
        "004": "15000_19999",
        "005": "20000_24999",
        "006": "25000_29999",
        "007": "30000_34999",
        "008": "35000_39999",
        "009": "40000_49999",
        "010": "50000_59999",
        "011": "60000_69999",
        "012": "70000_79999",
        "013": "80000_89999",
        "014": "90000_99999",
        "015": "100000_124999",
        "016": "125000_149999",
        "017": "150000_174999",
        "018": "175000_199999",
        "019": "200000_249999",
        "020": "250000_299999",
        "021": "300000_399999",
        "022": "400000_499999",
        "023": "500000_749999",
        "024": "750000_999999",
        "025": "1000000_1499999",
        "026": "1500000_1999999",
        "027": "2000000_more",
    })


@register
class MedianMonthlyHousingCostsDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "medianmonthlyhousingcosts"
    UNIVERSE = "occupied housing units with monthly housing costs"
    RAW_TABLE_NAME = "B25105"
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        "001": "median"
    })


@register
class MedianHousingValueDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "medianhousingvalue"
    UNIVERSE = "owner-occupied housing units"
    RAW_TABLE_NAME = 'B25077'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        "001": "median"
    })


@register
class MedianGrossRentDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "mediangrossrent"
    UNIVERSE = "renter-occupied housing units paying cash rent"
    RAW_TABLE_NAME = 'B25064'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        "001": "median"
    })


@register
class TenurePopDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "tenurepop"
    UNIVERSE = "total population in occupied housing units"
    RAW_TABLE_NAME = 'B25008'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': 'universe',
        '002': 'owner_occupied',
        '003': 'renter_occupied'
    })


@register
class TenureDownloader(BaseTableConfig):
    PROCESSED_TABLE_NAME = "tenure"
    UNIVERSE = "occupied housing units"
    RAW_TABLE_NAME = 'B25003'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': 'universe',
        '002': 'owner_occupied',
        '003': 'renter_occupied'
    })


@register
class TenureLatinoDownloader(TenureDownloader):
    PROCESSED_TABLE_NAME = "tenurelatino"
    RAW_TABLE_NAME = 'B25003I'


@register
class TenureWhiteDownloader(TenureDownloader):
    PROCESSED_TABLE_NAME = "tenurewhite"
    RAW_TABLE_NAME = 'B25003H'


@register
class TenureBlackDownloader(TenureDownloader):
    PROCESSED_TABLE_NAME = "tenureblack"
    RAW_TABLE_NAME = 'B25003B'


@register
class TenureAsianDownloader(TenureDownloader):
    PROCESSED_TABLE_NAME = "tenureasian"
    RAW_TABLE_NAME = 'B25003D'


@register
class MedianRentBedroomDownloader(TenureDownloader):
    PROCESSED_TABLE_NAME = "mediangrossrentbybedroom"
    UNIVERSE = " renter-occupied housing units paying cash rent"
    RAW_TABLE_NAME = 'B25031'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': 'median_gross_rent',
        '003': 'median_gross_rent_1_bedroom'
    })


@register
class GrossRentBedroomDownloader(TenureDownloader):
    PROCESSED_TABLE_NAME = "grossrentbybedroom"
    UNIVERSE = "renter occupied housing units"
    RAW_TABLE_NAME = 'B25068'
    """ these are all one bed with cash rent, for the aggregator"""
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '011': 'total_one_bed',
        '012': 'total_one_bed_with_cash_rent',
        '013': '0_299',
        '014': '300_499',
        '015': '500_749',
        '016': '750_999',
        '017': '1000_1499',
        '018': '1500_more',
    })



#! /usr/bin/env python
# -*- coding: utf-8 -*
# flake8: NOQA
import jinja2
import logging
import pathlib
from .tables import TABLE_LIST
logger = logging.getLogger(__name__)


def download_everything(*args, **kwargs):
    """
    Download all the data.
    """
    logger.debug("Downloading all datasets for all geographies")
    for klass in TABLE_LIST:
        obj = klass(*args, **kwargs)
        logger.debug(f"Downloading {klass.PROCESSED_TABLE_NAME} dataset")
        obj.download_everything()


def inventory_everything(docs_dir):
    """
    Create markdown files describing all of the data tables.
    """
    docs_path = pathlib.Path(docs_dir)
    if not docs_path.exists():
        docs_path.mkdir()
    template_loader = jinja2.FileSystemLoader([pathlib.Path(__file__).parent.joinpath("templates")])
    template_env = jinja2.Environment(loader=template_loader)
    list_template = template_env.get_template("table_list.md")
    with open(docs_path.joinpath("TABLES.md"), 'w') as f:
        f.write(list_template.render(object_list=TABLE_LIST))


__all__ = ("TABLE_LIST", "download_everything")



#! /usr/bin/env python
# -*- coding: utf-8 -*
"""
Handlers to process table configurations for each of the Census' different geography types.
"""
import time
import logging
import collections
import pandas as pd
from us import states
from census import Census
logger = logging.getLogger(__name__)

# Base downloader classes


class BaseGeoTypeDownloader(object):
    """
    Base class for downloading raw data from the Census API.

    Expects a TableConfig instance and year as input.
    """
    YEAR_LIST = [
        2018,
        2017,
        2016,
        2015,
        2014,
        2013,
        2012,
        2011,
        2010,
        2009
    ]

    def __init__(self, config, year):
        # Set the input variables
        self.config = config
        self.year = year

        # Connect to the Census API for the provided source
        self.api = getattr(Census(self.config.CENSUS_API_KEY, year=year), self.config.source)

        # Validate the year
        if self.year not in self.YEAR_LIST:
            raise NotImplementedError("Census API does not support this year for this geotype")

        # Validate the geotype
        valid_geotype_slugs = [gt.replace("_", "") for gt in self.config.GEOTYPE_LIST]
        if self.slug not in valid_geotype_slugs:
            raise NotImplementedError(f"Data only available for {', '.join(self.config.GEOTYPE_LIST)}")

        # Prepare raw field list
        self.api_fields = self.get_api_fields()

        # Set the CSV names
        self.raw_csv_name = f"{self.config.source}_{self.year}_{self.config.PROCESSED_TABLE_NAME}_{self.slug}.csv"
        self.raw_csv_path = self.config.raw_data_dir.joinpath(self.raw_csv_name)
        self.processed_csv_name = f"{self.config.source}_{self.year}_{self.config.PROCESSED_TABLE_NAME}_{self.slug}.csv"
        self.processed_csv_path = self.config.processed_data_dir.joinpath(self.processed_csv_name)

    @property
    def api_filter(self):
        """
        Returns the filter that will retrieve the correct data from the API.
        """
        return {'for': f'{self.raw_geotype}:*'}

    def get_api_fields(self):
        """
        Returns the fields to be fetched from the API.
        """
        return list(self.get_raw_field_map().keys())

    def get_raw_field_map(self):
        """
        Returns a crosswalk between the raw API fields and our processed humanized field names.
        """
        """
        API doesn't have margin of error for 2009, so we will need a different crosswalk for the 2009 data.
        """
        if self.year == 2009:
            field_map_2009 = collections.OrderedDict({
                'NAME': "name"
            })
            field_suffix_map_2009 = {
                "E": "",
                "M": "_moe"
            }
            for field_id, field_name in self.config.RAW_FIELD_CROSSWALK.items():
                for field_suffix, name_suffix in field_suffix_map_2009.items():
                    full_raw_id = f"{self.config.RAW_TABLE_NAME}_{field_id}{field_suffix}"
                    processed_name = f"{field_name}{name_suffix}".strip()
                    field_map_2009[full_raw_id] = processed_name
            return field_map_2009
        else:
            field_map = collections.OrderedDict({
                'NAME': "name"
            })
            field_suffix_map = {
                "E": "",
                "EA": "_annotation",
                "M": "_moe",
                "MA": "_moe_annotation"
            }
            for field_id, field_name in self.config.RAW_FIELD_CROSSWALK.items():
                for field_suffix, name_suffix in field_suffix_map.items():
                    full_raw_id = f"{self.config.RAW_TABLE_NAME}_{field_id}{field_suffix}"
                    processed_name = f"{field_name}{name_suffix}".strip()
                    field_map[full_raw_id] = processed_name
        return field_map

    def get_raw_data(self):
        """
        Returns the data we want from the API.
        """
        logger.debug(f"Downloading {self.slug} {self.config.PROCESSED_TABLE_NAME} data from raw {self.config.RAW_TABLE_NAME} table in {self.year} {self.config.source} count")
        # Get the raw data
        return self.api.get(self.api_fields, self.api_filter)

    def create_geoid(self, row):
        """
        Create the unique identifier we prefer for this geotype.
        """
        return row[self.raw_geotype]

    def download(self):
        """
        Downloads raw data from the Census API.
        Returns path to CSV.
        """
        # Skip hitting the API if we've already got the file, as long as we're not forcing it.
        if self.raw_csv_path.exists() and not self.config.force:
            logger.debug(f"Raw CSV already exists at {self.raw_csv_path}")
            return self.raw_csv_path

        # Get the data
        data = self.get_raw_data()

        # Convert it to a dataframe
        df = pd.DataFrame.from_records(data)

        # Write it to disk
        logger.debug(f"Writing raw ACS table to {self.raw_csv_path}")
        df.to_csv(self.raw_csv_path, index=False, encoding="utf-8")

        # Pause to be polite to the API.
        time.sleep(0.5)

        # Hand the path back
        return self.raw_csv_path

    def process(self):
        """
        Converts the raw file to be used by humans.

        Returns path to CSV.
        """
        # Figure out where to write it
        if self.processed_csv_path.exists() and not self.config.force:
            logger.debug(f"Processed CSV already exists at {self.processed_csv_path}")
            return self.processed_csv_path

        # Read in the raw CSV of data from the ACS
        df = pd.read_csv(self.raw_csv_path, dtype=str)

        # Get the crosswalk between raw and processed field names
        field_name_mapper = self.get_raw_field_map()

        # Cast numbers to floats
        for field in field_name_mapper.keys():
            if "_" in field and (field.endswith("E") or field.endswith("M")):
                df[field] = df[field].astype(pd.np.float64)

        # Rename fields with humanized names
        df.rename(columns=field_name_mapper, inplace=True)

        # Add a combined GEOID column with a Census unique identifer
        df['geoid'] = df.apply(self.create_geoid, axis=1)

        if hasattr(self.config, 'process'):
            df = self.config.process(df)

        # Write it out
        logger.debug(f"Writing CSV to {self.processed_csv_path}")
        df.set_index(["geoid", "name"]).to_csv(
            self.processed_csv_path,
            index=True,
            encoding="utf-8"
        )
        return self.processed_csv_path


class BaseStateLevelGeoTypeDownloader(BaseGeoTypeDownloader):
    """
    Download and stitch together raw data the Census API only provides state by state.
    """
    def get_api_filter(self, state):
        """
        Returns an API filter to retrieve the correct data for the provided state.
        """
        return {
            'for': f'{self.raw_geotype}:*',
            'in': f'state: {state.fips}'
        }

    def get_raw_data(self):
        """
        Returns the data we want from the API.
        """
        api_data = []
        for state in states.STATES:
            logger.debug(f"Downloading {self.slug} {self.config.PROCESSED_TABLE_NAME} data from the raw {self.config.RAW_TABLE_NAME} table in {state} for the {self.year} {self.config.source} count")
            # Get the raw data
            api_filter = self.get_api_filter(state)
            state_data = self.api.get(self.api_fields, api_filter)
            api_data.extend(state_data)
        return api_data

    def create_geoid(self, row):
        return row['state'] + row[self.raw_geotype]

#
# Geotype downloaders
#


class NationwideDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the nationwide level.
    """
    slug = "nationwide"
    raw_geotype = "us"

    @property
    def api_filter(self):
        return {'for': 'us:1'}

    def create_geoid(self, row):
        return 1


class RegionsDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the region level.
    """
    slug = "regions"
    raw_geotype = "region"


class DivisionsDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the division level.
    """
    slug = "divisions"
    raw_geotype = "division"


class StatesDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the state level.
    """
    slug = "states"
    raw_geotype = "state"


class CongressionalDistrictsDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the congressional-district level.
    """
    slug = "congressionaldistricts"
    raw_geotype = "congressional district"

    def create_geoid(self, row):
        return row['state'] + row[self.raw_geotype]


class StateLegislativeUpperDistrictsDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the upper-level state-legislative-district level.
    """
    slug = "statelegislativeupperdistricts"
    raw_geotype = "state legislative district (upper chamber)"


class StateLegislativeLowerDistrictsDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the lower-level state-legislative-district level.
    """
    slug = "statelegislativelowerdistricts"
    raw_geotype = "state legislative district (lower chamber)"


class CountiesDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the county level.
    """
    slug = "counties"
    raw_geotype = "county"

    def create_geoid(self, row):
        return row['state'] + row[self.raw_geotype]


class PlacesDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the place level.
    """
    slug = "places"
    raw_geotype = "place"

    def create_geoid(self, row):
        return row['state'] + row[self.raw_geotype]


class UrbanAreasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the urban-area level.
    """
    slug = "urbanareas"
    raw_geotype = "urban area"


class MsasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the metropolitan-statistical-area level.
    """
    slug = "msas"
    raw_geotype = "metropolitan statistical area/micropolitan statistical area"


class CsasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the combined-statistical-area level.
    """
    slug = "csas"
    raw_geotype = "combined statistical area"


class PumasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the public-use-microdata level.
    """
    slug = "pumas"
    raw_geotype = "public use microdata area"


class NectasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the New England city and town level.
    """
    slug = "nectas"
    raw_geotype = "new england city and town area"


class CnectasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the combined New England city and town level.
    """
    YEAR_LIST = [
        2018,
        2017,
        2016,
        2015,
        2014,
        2013,
        2012,
        2011,
        2010
    ]

    slug = "cnectas"
    raw_geotype = "combined new england city and town area"


class AiannhHomelandsDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the Native American homeland level.
    """
    slug = "aiannhhomelands"
    raw_geotype = "american indian area/alaska native area/hawaiian home land"


class TractsDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the tract level.
    """
    slug = "tracts"
    raw_geotype = "tract"

    def create_geoid(self, row):
        return row['state'] + row['county'] + row[self.raw_geotype]


class ZctasDownloader(BaseGeoTypeDownloader):
    """
    Download raw data at the zipcode-tabulation-area level.
    """
    YEAR_LIST = [
        2018,
        2017,
        2016,
        2015,
        2014,
        2013,
        2012,
        2011
    ]

    slug = "zctas"
    raw_geotype = "zip code tabulation area"


class UnifiedSchoolDistrictsDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the unified school district level.
    """
    slug = "unifiedschooldistricts"
    raw_geotype = "school district (unified)"


class ElementarySchoolDistrictsDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the elementary school district level.
    """
    slug = "elementaryschooldistricts"
    raw_geotype = "school district (elementary)"


class SecondarySchoolDistrictsDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the secondary school district level.
    """
    slug = "secondaryschooldistricts"
    raw_geotype = "school district (secondary)"


class AlaskaNativeDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the Alaska Native regional corporation level.
    """
    slug = "alaskanative"
    raw_geotype = "alaska native regional corporation"


class CountySubdivisionDownloader(BaseStateLevelGeoTypeDownloader):
    """
    Download raw data at the county subdivision level.
    """
    slug = "countysubdivision"
    raw_geotype = "county subdivision"



#! /usr/bin/env python
# -*- coding: utf-8 -*
import collections
from census_data_downloader.core.tables import BaseTableConfig
from census_data_downloader.core.decorators import register


@register
class PovertyDownloader(BaseTableConfig):
    """
    A simplified version of the poverty table that only returns grand totals.
    """
    PROCESSED_TABLE_NAME = "poverty"
    UNIVERSE = "population for whom poverty status is determined"
    RAW_TABLE_NAME = 'B17001'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': 'universe',
        '002': 'income_past12months_below_poverty_level',
        '031': 'income_past12months_at_or_above_poverty_level'
    })


@register
class PovertyBySexDownloader(BaseTableConfig):
    """
    A simplified version of the poverty table that only returns totals by sex.
    """
    PROCESSED_TABLE_NAME = "povertystatusbygender"
    UNIVERSE = "population for whom poverty status is determined"
    RAW_TABLE_NAME = 'B17001'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': "total",
        '002': "total_below_poverty_level",
        '003': "male_below_poverty_level",
        '017': "female_below_poverty_level",
        '031': "total_at_or_above_poverty_level",
        '032': "male_at_or_above_poverty_level",
        '046': "female_at_or_above_poverty_level"
    })


@register
class PovertyAgeDownloader(BaseTableConfig):
    """
    The full table.
    """
    PROCESSED_TABLE_NAME = "povertystatusbyage"
    UNIVERSE = "population for whom poverty status is determined"
    RAW_TABLE_NAME = 'B17001'
    RAW_FIELD_CROSSWALK = collections.OrderedDict({
        '001': "total",
        '002': "total_below_poverty_level",
        '003': "male_below_poverty_level",
        '004': "male_under_5_below_poverty_level",
        '005': "male_5_below_poverty_level",
        '006': "male_6_to_11_below_poverty_level",
        '007': "male_12_to_14_below_poverty_level",
        '008': "male_15_below_poverty_level",
        '009': "male_16_to_17_below_poverty_level",
        '010': "male_18_to_24_below_poverty_level",
        '011': "male_25_to_34_below_poverty_level",
        '012': "male_35_to_44_below_poverty_level",
        '013': "male_45_to_54_below_poverty_level",
        '014': "male_55_to_64_below_poverty_level",
        '015': "male_65_to_74_below_poverty_level",
        '016': "male_75_and_over_below_poverty_level",
        '017': "female_below_poverty_level",
        '018': "female_under_5_below_poverty_level",
        '019': "female_5_below_poverty_level",
        '020': "female_6_to_11_below_poverty_level",
        '021': "female_12_to_14_below_poverty_level",
        '022': "female_15_below_poverty_level",
        '023': "female_16_to_17_below_poverty_level",
        '024': "female_18_to_24_below_poverty_level",
        '025': "female_25_to_34_below_poverty_level",
        '026': "female_35_to_44_below_poverty_level",
        '027': "female_45_to_54_below_poverty_level",
        '028': "female_55_to_64_below_poverty_level",
        '029': "female_65_to_74_below_poverty_level",
        '030': "female_75_and_over_below_poverty_level",
        '031': "total_at_or_above_poverty_level",
        '032': "male_at_or_above_poverty_level",
        '033': "male_under_5_at_or_above_poverty_level",
        '034': "male_5_at_or_above_poverty_level",
        '035': "male_6_to_11_at_or_above_poverty_level",
        '036': "male_12_to_14_at_or_above_poverty_level",
        '037': "male_15_at_or_above_poverty_level",
        '038': "male_16_to_17_at_or_above_poverty_level",
        '039': "male_18_to_24_at_or_above_poverty_level",
        '040': "male_25_to_34_at_or_above_poverty_level",
        '041': "male_35_to_44_at_or_above_poverty_level",
        '042': "male_45_to_54_at_or_above_poverty_level",
        '043': "male_55_to_64_at_or_above_poverty_level",
        '044': "male_65_to_74_at_or_above_poverty_level",
        '045': "male_75_and_over_at_or_above_poverty_level",
        '046': "female_at_or_above_poverty_level",
        '047': "female_under_5_at_or_above_poverty_level",
        '048': "female_5_at_or_above_poverty_level",
        '049': "female_6_to_11_at_or_above_poverty_level",
        '050': "female_12_to_14_at_or_above_poverty_level",
        '051': "female_15_at_or_above_poverty_level",
        '052': "female_16_to_17_at_or_above_poverty_level",
        '053': "female_18_to_24_at_or_above_poverty_level",
        '054': "female_25_to_34_at_or_above_poverty_level",
        '055': "female_35_to_44_at_or_above_poverty_level",
        '056': "female_45_to_54_at_or_above_poverty_level",
        '057': "female_55_to_64_at_or_above_poverty_level",
        '058': "female_65_to_74_at_or_above_poverty_level",
        '059': "female_75_and_over_at_or_above_poverty_level"
    })

    def process(self, df):

        # Calculate totals for both genders together
        groups = [
            'under_5_below_poverty_level',
            '5_below_poverty_level',
            '6_to_11_below_poverty_level',
            '12_to_14_below_poverty_level',
            '15_below_poverty_level',
            '16_to_17_below_poverty_level',
            '18_to_24_below_poverty_level',
            '25_to_34_below_poverty_level',
            '35_to_44_below_poverty_level',
            '45_to_54_below_poverty_level',
            '55_to_64_below_poverty_level',
            '65_to_74_below_poverty_level',
            '75_and_over_below_poverty_level',
            'under_5_at_or_above_poverty_level',
            '5_at_or_above_poverty_level',
            '6_to_11_at_or_above_poverty_level',
            '12_to_14_at_or_above_poverty_level',
            '15_at_or_above_poverty_level',
            '16_to_17_at_or_above_poverty_level',
            '18_to_24_at_or_above_poverty_level',
            '25_to_34_at_or_above_poverty_level',
            '35_to_44_at_or_above_poverty_level',
            '45_to_54_at_or_above_poverty_level',
            '55_to_64_at_or_above_poverty_level',
            '65_to_74_at_or_above_poverty_level',
            '75_and_over_at_or_above_poverty_level'
        ]
        for g in groups:
            df[f'total_{g}'] = df[f'male_{g}'] + df[f'female_{g}']

        # Calculate custom group sets
        groupsets = collections.OrderedDict({
            "0_to_17_total": [
                'under_5_below_poverty_level',
                '5_below_poverty_level',
                '6_to_11_below_poverty_level',
                '12_to_14_below_poverty_level',
                '15_below_poverty_level',
                '16_to_17_below_poverty_level',
                'under_5_at_or_above_poverty_level',
                '5_at_or_above_poverty_level',
                '6_to_11_at_or_above_poverty_level',
                '12_to_14_at_or_above_poverty_level',
                '15_at_or_above_poverty_level',
                '16_to_17_at_or_above_poverty_level',
            ],
            "0_to_17_below_poverty_level": [
                'under_5_below_poverty_level',
                '5_below_poverty_level',
                '6_to_11_below_poverty_level',
                '12_to_14_below_poverty_level',
                '15_below_poverty_level',
                '16_to_17_below_poverty_level',
            ],
            "18_to_34_total": [
                '18_to_24_below_poverty_level',
                '25_to_34_below_poverty_level',
                '18_to_24_at_or_above_poverty_level',
                '25_to_34_at_or_above_poverty_level',
            ],
            "18_to_34_below_poverty_level": [
                '18_to_24_below_poverty_level',
                '25_to_34_below_poverty_level',
            ],
            "35_to_64_total": [
                '35_to_44_below_poverty_level',
                '45_to_54_below_poverty_level',
                '55_to_64_below_poverty_level',
                '35_to_44_at_or_above_poverty_level',
                '45_to_54_at_or_above_poverty_level',
                '55_to_64_at_or_above_poverty_level',
            ],
            "35_to_64_below_poverty_level": [
                '35_to_44_below_poverty_level',
                '45_to_54_below_poverty_level',
                '55_to_64_below_poverty_level',
            ],
            "65_and_over_total": [
                '65_to_74_below_poverty_level',
                '75_and_over_below_poverty_level',
                '65_to_74_at_or_above_poverty_level',
                '75_and_over_at_or_above_poverty_level',
            ],
            "65_and_over_below_poverty_level": [
                '65_to_74_below_poverty_level',
                '75_and_over_below_poverty_level',
            ],
            "0_to_11_total": [
                '12_to_14_below_poverty_level',
                '15_below_poverty_level',
                '16_to_17_below_poverty_level',
                '12_to_14_at_or_above_poverty_level',
                '15_at_or_above_poverty_level',
                '16_to_17_at_or_above_poverty_level',
            ],
            "0_to_11_below_poverty_level": [
                '12_to_14_below_poverty_level',
                '15_below_poverty_level',
                '16_to_17_below_poverty_level',
            ],
            "12_to_17_total": [
                '12_to_14_below_poverty_level',
                '15_below_poverty_level',
                '16_to_17_below_poverty_level',
                '12_to_14_at_or_above_poverty_level',
                '15_at_or_above_poverty_level',
                '16_to_17_at_or_above_poverty_level',
            ],
            "12_to_17_below_poverty_level": [
                '12_to_14_below_poverty_level',
                '15_below_poverty_level',
                '16_to_17_below_poverty_level',
            ],
            "18_to_64_total": [
                '18_to_24_below_poverty_level',
                '25_to_34_below_poverty_level',
                '35_to_44_below_poverty_level',
                '45_to_54_below_poverty_level',
                '55_to_64_below_poverty_level',
                '18_to_24_at_or_above_poverty_level',
                '25_to_34_at_or_above_poverty_level',
                '35_to_44_at_or_above_poverty_level',
                '45_to_54_at_or_above_poverty_level',
                '55_to_64_at_or_above_poverty_level',
            ],
            "18_to_64_below_poverty_level": [
                '18_to_24_below_poverty_level',
                '25_to_34_below_poverty_level',
                '35_to_44_below_poverty_level',
                '45_to_54_below_poverty_level',
                '55_to_64_below_poverty_level',
            ]
        })
        for groupset, group_list in groupsets.items():
            df[f'total_{groupset}'] = df[[f'total_{f}' for f in group_list]].sum(axis=1)
            df[f'male_{groupset}'] = df[[f'male_{f}' for f in group_list]].sum(axis=1)
            df[f'female_{groupset}'] = df[[f'female_{f}' for f in group_list]].sum(axis=1)

        # Pass it back
        return df


@register
class PovertyLatinoDownloader(PovertyAgeDownloader):
    PROCESSED_TABLE_NAME = "povertylatino"
    RAW_TABLE_NAME = 'B17001I'
    UNIVERSE = "Latino population for whom poverty status is determined"


@register
class PovertyWhiteDownloader(PovertyAgeDownloader):
    PROCESSED_TABLE_NAME = "povertywhite"
    RAW_TABLE_NAME = 'B17001A'
    UNIVERSE = "white population for whom poverty status is determined"


@register
class PovertyBlackDownloader(PovertyAgeDownloader):
    PROCESSED_TABLE_NAME = "povertyblack"
    RAW_TABLE_NAME = 'B17001B'
    UNIVERSE = "Black population for whom poverty status is determined"


@register
class PovertyAsianDownloader(PovertyAgeDownloader):
    PROCESSED_TABLE_NAME = "povertyasian"
    RAW_TABLE_NAME = 'B17001D'
    UNIVERSE = "Asian population for whom poverty status is determined"
